#!/usr/bin/python3
# -*- coding: UTF-8 -*-
from collections import OrderedDict

import re
from s3200 import core
import logging


logger = logging.getLogger('s3200')
# TODO remove dummy integrate into unittests dummy expected input -> expected output

# noinspection PyPep8Naming,PyPep8Naming
class DummySerial(object):
    """ A dummy serial port implementation that behaves like a simplified connected s3200 """

    COMM_LIST = OrderedDict({
        # value working hours answer 55
        '02 FD 00 03 30 00 62 F2': b'\x02\xfd\x00\x030\x007\r',
        # simulate transmission error wrong checksum when asked for 00 59 heater_circuit_18_is
        '02 FD 00 03 30 00 59 BF': b'\x02\xfd\x00\x02\x000\x10h',
        # other values answer 4322
        '02 FD .. .. 30 (?!(00 62 F2))(?!(00 59 BF)).*': b'\x02\xFD\x00\x03\x30\x10\xE2\x42',
        # answer to 22 test
        '02 FD .. .. 22 .*': 'return',
        # answer to 41 sw version and current datetime
        '02 FD .. .. 41 .*': b'\x02\xFD\x00\x0C\x41\x50\x04\x04\x14\x00\x1F\x12\x15\x0B\x07\x0A\x38',
        # error buffer get_error
        '02 FD .. .. 47 .*': b'\x02\xfd\x009G\x01\x00m\xa2\x04%&\n\x0b\x04\rZ\xfcndversuch nicht gelungen von Hand Anheizen!\x92',
                                #b'\x02\xfd\x009G\x01\x00m\xa3\x01\x02\x00\xfe\x12\x0c\x04\x02\x00\x0c' +
                             #b'Z\xfcndversuch nicht gelungen von Hand Anheizen!\xbc',
        # error buffer get_next_error EMPTY bytearray(b'\x02\xfd\x00\x01G\xcb')
        '02 FD .. .. 48 .*': b'\x02\xfd\x00\x02\x00G\x00\xce',
        '02 FD .. 02 00 48 .*': b'\x02\xfd\x00\x02\x00G\x00\xce',
                            #b'\x02\xFD\x00\x02\x47\xCE',
                             #b'\x02\xFD\x00\x02\x00\x47\x00\xCE'
        #get configuration
        '02 FD .. .. 40 .*': b'\x02\xFD\x00\x59\x40\x00\x00\x00\x04\x00\x24\x00\x03\x00\x00\x00' +
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00' +
                             b'\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00' +
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00' +
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' +
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x49',
        #get_heater_state_and_mode
        '02 FD .. .. 51 .*': b'\x02\xFD\x00\x18\x51\x02\x00\x00\xDC\x62\x65\x72\x67\x61\x6E\x67' +
                             b'\x73\x62\x65\x74\x72\x3B\x53\x54\xD6\x52\x55\x4E\x47\x8C',
        #menu item start
        '02 FD .. .. 37 .*': b'\x02\xFD\x00\x44\x37\x01\x07\x00\x01\x72\x00\x00\x04\x00\x00\x00'
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00\x53'
                             b'\x00\xF7\x50\x72\x6F\x70\x6F\x72\x74\x69\x6F\x6E\x61\x6C\x66\x61'
                             b'\x6B\x74\x6F\x72\x20\x64\x65\x73\x20\x4D\x69\x73\x63\x68\x65\x72'
                             b'\x72\x65\x67\x6C\x65\x72\x73\x00\x7D',

        #menu next item empty
        '02 FD .. .. 38 .*': b'\x02\xfd\x00\x02\x008\x00\x4f',
        '02 FD .. 02 00 38 .*': b'\x02\xfd\x00\x02\x008\x00\x4f',
        #get setting
        '02 FD .. .. 55 .*': b'\x02\xFD\x00\xFE\x14\x55\x00\x00\x1C\xB0\x00\x00\x02\x00\x00\xA8'
                             b'\x00\x46\x00\x5A\x00\x50\x00\x00\x00\x15',
        #set setting out of range 120
        '02 FD 00 05 39 00 1C 00 F0 71': b'\x02\xfd\x00\x059\x00\x1c\x00x\xe9', #02 FD 00 05 39 00 1C 00 A2 87
        #set setting right
        '02 FD .. .. 39 .*': 'return*2',
        #available value
        '02 FD .. .. 31 .*': b'\x02\xFD\x00\x1B\x31\x01\x00\x02\x00\x00\x02\x00\x00\xB0\x00\x00'
                             b'\x4B\x65\x73\x73\x65\x6C\x74\x65\x6D\x70\x65\x72\x61\x74\x75\x72'
                             b'\x00\xDE',
        #available value next empty
        '02 FD .. .. 32 .*': b'\x02\xfd\x00\x02\x001\x00\x54',
        '02 FD .. 02 00 32 .*': b'\x02\xfd\x00\x02\x001\x00\x54',
        # digital input
        '02 FD .. .. 46 .*': b'\x02\xfd\x00\x03FA\x01\x0e',
        # digital output
        '02 FD .. .. 44 .*': b'\x02\xfd\x00\x03DA\x01\x08',
        # analog output
        '02 FD .. .. 45 .*': b'\x02\xfd\x00\x03E\xffco',

    })

    def __init__(self, port=None, baud_rate=None, eight_bits=None, parity=None, stop_bits=None, timeout=None):
        self.in_buffer = bytearray()
        self.out_buffer = bytearray()

    def flushInput(self):
        self.do_processing()

    def close(self):
        self.in_buffer = bytearray() #  .clear()
        self.out_buffer = bytearray() #  .clear()

    def write(self, write_bytes: bytes):
        for b in write_bytes:
            self.out_buffer.append(b)
        self.do_processing()

    def read(self, length=1):
        """dummy read function
        :param length: bytes to read
        """

        ret = self.in_buffer[0:length]
        self.in_buffer = self.in_buffer[length:]

        if len(ret) < length:
            raise core.NothingToReadError("Dummy has not enough bytes.")
        return ret

    def do_processing(self):
        for key, value in DummySerial.COMM_LIST.items():
            r = re.compile(key)
            hex_string = core.convert_bytes_to_hex(bytes(self.out_buffer)).upper()
            if r.match(hex_string):

                if value == 'return':
                    value = bytes(self.out_buffer)
                if value == 'return*2':
                    value = bytes(self.out_buffer) + bytes(self.out_buffer)

                self.out_buffer = bytearray() #  .clear()

                for i in value:
                    self.in_buffer.append(i)
                logger.info("Dummy in: {0} out: {1}".format(str(hex_string), str(value)))
                return
        raise NotImplementedError("out_buff: " + core.convert_bytes_to_hex(self.out_buffer) +
                                  " in_buff: " + str(self.in_buffer))

    def inWaiting(self):
        return len(self.in_buffer)




#!/usr/bin/python3
# -*- coding: UTF-8 -*-
from unittest import TestCase
from s3200.core import CommunicationError
from s3200.net import Frame, Connection


class TestConnection(TestCase):

    def setUp(self):
        self.c = Connection('dummy')

    def test_send_request(self):

        # Send normal request working hours
        f = Frame(b'\x30', b'\x00\x62')
        return_frame = self.c.send_frame(f)

        self.assertEquals(return_frame.command, b'\x30')
        self.assertEquals(return_frame.payload, b'\x00\x37')

        # Test what does happen when a wrong checksum error occurs
        f = Frame(b'\x30', b'\x00\x59')
        # print('Send:'+str(f))
        # print('Should be:'+str(Frame.from_bytes(b'\x02\xfd\x00\x02\x000\x10h')))
        # return_value = c.send_request(f)
        # print('Is:'+return_value)

        self.assertRaises(CommunicationError, self.c.send_frame, f)



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

from struct import Struct
from collections import OrderedDict

#---STATIC VARIABLES---
START_BYTES = b'\x02\xFD'

#List of bytes to escape. Reverse key and value, to get unescape list
ESCAPE_LIST = OrderedDict(
    [
        (b'\x2B', b'\x2B\x00'),
        (b'\xFE', b'\xFE\x00'),
        (b'\x02', b'\x02\x00'),
        (b'\x11', b'\xFE\x12'),
        (b'\x13', b'\xFE\x14'),
    ]
)
UNESCAPE_LIST = OrderedDict((v, k) for k, v in ESCAPE_LIST.items())

#The bytes which identify that something is escaped
ESCAPED_IDENTIFIER = bytes([0x02, 0xFE, 0x2B])

#Structs
StructShort = Struct('!h')
StructDateDayTime = Struct('!7b')
StructDateTime = Struct('!6b')

#### COMMAND AND VALUE DEFINITIONS ####

COMMAND_DEFINITIONS = {
    'test_connection':   {
        'address': b'\x22',
        'description': 'Returns the given frame. For testing the connection',
    },
    'get_value':   {
        'address': b'\x30',
        'description': 'Read a value from the heater.',
    },
    'get_available_value':   {
        'address': b'\x31',
        'description': 'Get the first value from the available values list',
    },
    'get_next_available_value':   {
        'address': b'\x32',
        'description': 'Get the next value from the available values list',
    },
    'get_menu_item':   {
        'address': b'\x37',
        'description': 'Gets the first menu item.',
    },
    'get_next_menu_item':   {
        'address': b'\x38',
        'description': 'Gets the next menu item.',
    },
    'set_setting':   {
        'address': b'\x39',
        'description': 'Set a specific setting',
    },
    'get_configuration':   {
        'address': b'\x40',
        'description': 'Gets the configuration from the heater. eg which boilers are connected',
    },
    'get_version_and_datetime':   {
        'address': b'\x41',
        'description': 'Get Version, Date and Time',
    },
    'get_time_slot':   {
        'address': b'\x42',
        'description': 'Get the first time slot',
    },
    'get_next_time_slot':   {
        'address': b'\x43',
        'description': 'Get the next time slot',
    },
    'get_digital_output':   {
        'address': b'\x44',
        'description': 'Gets the value of a digital (True, False) output.',
    },
    'get_analog_output':   {
        'address': b'\x45',
        'description': 'Gets the value of a analog output (0-100%)',
    },
    'get_digital_input':   {
        'address': b'\x46',
        'description': 'Gets the value of a digital (True, False) output.',
    },
    'get_error':   {
        'address': b'\x47',
        'description': 'Gets the first error from the error list',
    },
    'get_next_error':   {
        'address': b'\x48',
        'description': 'Gets the next error from the error list',
    },
    'set_time_slot':   {
        'address': b'\x50',
        'description': '',
    },
    'get_heater_state_and_mode':   {
        'address': b'\x51',
        'description': 'Gets state and mode from the heater',
    },
    'set_datetime':   {
        'address': b'\x54',
        'description': 'Set the date and time of the heater',
    },
    'get_setting':   {
        'address': b'\x55',
        'description': 'Get the current value and min, max value of the setting',
    },
    'manipulate_digital_output':   {
        'address': b'\x58',
        'description': '',
    },
    'manipulate_analog_output':   {
        'address': b'\x59',
        'description': '',
    },
    'manipulate_digital_input':   {
        'address': b'\x5A',
        'description': '',
    },
    'get_force':   {
        'address': b'\x5E',
        'description': '',
    },
    'set_force':   {
        'address': b'\x7E',
        'description': '',
    },

}

#The address of the actual values
VALUE_DEFINITIONS = OrderedDict({
    'heating_boiler_temperature':   {'address': b'\x00\x00', 'factor': 2, 'local_name': 'Heizkesseltemperatur'},
    'exhaust_temperature':          {'address': b'\x00\x01', 'factor': 1, 'local_name': 'Abgastemperatur'},
    'board_temperature':            {'address': b'\x00\x02', 'factor': 2, 'local_name': 'Boardtemperatur'},
    'residual_oxygen':              {'address': b'\x00\x03', 'factor': 10, 'local_name': 'Restsauerstoff'},
    'outside_temperature':          {'address': b'\x00\x04', 'factor': 2, 'local_name': 'Außentemperatur'},
    'primary_air_inlet_position':   {'address': b'\x00\x05', 'factor': 1, 'local_name': 'Primärluftposition'},
    'secondary_air_inlet_position': {'address': b'\x00\x06', 'factor': 1, 'local_name': 'Sekundärluftposition'},
    'air_fan_rotation_speed':       {'address': b'\x00\x07', 'factor': 1, 'local_name': 'Saugzugdrehzahl'},

    'heater_circuit_1_is_temperature':      {'address': b'\x00\x15', 'factor': 2, 'local_name': 'Heizkreis 1 Ist'},
    'heater_circuit_1_should_temperature':  {'address': b'\x00\x16', 'factor': 2, 'local_name': 'Heizkreis 1 Soll'},

    'boiler_1_temperature':              {'address': b'\x00\x5d', 'factor': 2, 'local_name': 'Boilertemperatur 1'},
    'operating_hours':                   {'address': b'\x00\x62', 'factor': 1, 'local_name': 'Betriebsstunden'},
    'operating_hours_fire_preservation': {
        'address': b'\x00\x73',
        'factor': 1,
        'local_name': 'Betriebsstunden Feuererhaltung'
    },

    'buffer_1_top_temperature':     {'address': b'\x00\x76', 'factor': 2, 'local_name': 'Puffer 1 oben'},
    'buffer_1_middle_temperature':  {'address': b'\x00\x77', 'factor': 2, 'local_name': 'Puffer 1 mitte'},
    'buffer_1_bottom_temperature':  {'address': b'\x00\x78', 'factor': 2, 'local_name': 'Puffer 1 unten'},

    'buffer_1_pump':    {'address': b'\x00\x8c', 'factor': 1, 'local_name': 'Pufferpumpenansteuerung 1'},
    'boiler_1_pump':    {'address': b'\x00\x90', 'factor': 1, 'local_name': 'Boilerpumpenansteuerung 1'},
})

#The address of the settings
SETTING_DEFINITIONS = OrderedDict({
    'heating_boiler_should_temperature':   {'address': b'\x00\x1C', 'factor': 2, 'local_name': 'Kessel-Solltemperatur'},
    'start_firing':   {'address': b'\x02\x52', 'factor': 1, 'local_name': 'Zündung starten'},

})

DIGITAL_INPUT_DEFINITIONS = {
    'door_contact': b'\x00\x00',
    'stb': b'\x00\x01',
    'emergency_off': b'\x00\x01',
    'heating_boiler_release':b'\x00\x1F'
}
DIGITAL_OUTPUT_DEFINITIONS = {
    'heating_circuit_pump_1': b'\x00\x00',
    'heating_circuit_pump_2': b'\x00\x01',
    #...
}
ANALOG_OUTPUT_DEFINITIONS = {
    'primary_air': b'\x00\x00',
    'secondary_air': b'\x00\x01',
    #...
}

TIME_SLOT_DEFINITIONS = {
    'boiler_1_monday': b'\x00',
    'boiler_1_tuesday': b'\x01',
    'boiler_1_wednesday': b'\x02',
    'boiler_1_thursday': b'\x03',
    'boiler_1_friday': b'\x04',
    'boiler_1_saturday': b'\x05',
    'boiler_1_sunday': b'\x06',
}


#### MAPPINGS ####

ERROR_STATE = {
    1: 'New',
    2: 'Quittiert',
    4: 'Gone',
}
ERROR_STATE_LOCAL = {
    1: 'Gekommen',
    2: 'Quittiert',
    4: 'Gegangen',

}
TIME_SLOT_REVERSED = OrderedDict((v, k) for k, v in TIME_SLOT_DEFINITIONS.items())

#### STRUCTURES ####


CONFIGURATION_STRUCTURE = {
    'boiler_1': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 7},
    'boiler_2': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 6},
    'boiler_3': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 5},
    'boiler_4': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 4},
    'boiler_5': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 3},
    'boiler_6': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 2},
    'boiler_7': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 1},
    'boiler_8': {'start': 31, 'end': 32, 'type': 'flag', 'bit': 0},


    'heater_circuit_1': {'start': 37, 'end': 40, 'type': 'flag', 'bit': 23},
    'heater_circuit_2': {'start': 37, 'end': 40, 'type': 'flag', 'bit': 22},

    'remote_control_1': {'start': 45, 'end': 48, 'type': 'flag', 'bit': 23},
    'remote_control_2': {'start': 45, 'end': 48, 'type': 'flag', 'bit': 22},

    'solar_1': {'start': 63, 'end': 64, 'type': 'flag', 'bit': 7},
    'solar_2': {'start': 63, 'end': 64, 'type': 'flag', 'bit': 6},
     # TODO local_name needed???
}
MENU_ITEM_STRUCTURE = {
    'address': {'start': 25, 'end': 27, 'type': 'bytes'},
    'text': {'start': 29, 'end': -1, 'type': 'string'},
}

STATE_AND_MODE_STRUCTURE = {
    'text': {'start': 2, 'end': None, 'type': 'string'},
}

SETTING_STRUCTURE = {
    'address': {'start': 1, 'end': 3, 'type': 'bytes'},
    'unit': {'start': 3, 'end': 4, 'type': 'string'},
    'comma': {'start': 4, 'end': 5, 'type': 'short'},
    'factor': {'start': 6, 'end': 7, 'type': 'short'},
    'value': {'start': 7, 'end': 9, 'type': 'short'},
    'min_value': {'start': 9, 'end': 11, 'type': 'short'},
    'max_value': {'start': 11, 'end': 13, 'type': 'short'},
    'standard': {'start': 13, 'end': 15, 'type': 'short'},
}
DIGITAL_INPUT_STRUCTURE = {
    'mode': {'start': 0, 'end': 1, 'type': 'string'},
    'value': {'start': 1, 'end': 2, 'type': 'short'},
}
DIGITAL_OUTPUT_STRUCTURE = {
    'mode': {'start': 0, 'end': 1, 'type': 'string'},
    'value': {'start': 1, 'end': 2, 'type': 'short'},
}
ANALOG_OUTPUT_STRUCTURE = {
    'mode': {'start': 0, 'end': 1, 'type': 'short'},
    'value': {'start': 1, 'end': 2, 'type': 'short'},
}

TIME_SLOT_STRUCTURE = {
    #'address': {'start': 2, 'end': 3, 'type': 'bytes'},
    'address': {'start': 2, 'end': 3, 'type': 'bytes'},
    'name': {'start': 2, 'end': 3, 'type': 'bytes', 'reference': TIME_SLOT_REVERSED},
    'time_slot_1_start': {'start': 3, 'end': 4, 'type': 'time10'},
    'time_slot_1_end': {'start': 4, 'end': 5, 'type': 'time10'},
    'time_slot_2_start': {'start': 5, 'end': 6, 'type': 'time10'},
    'time_slot_2_end': {'start': 6, 'end': 7, 'type': 'time10'},
    'time_slot_3_start': {'start': 7, 'end': 8, 'type': 'time10'},
    'time_slot_3_end': {'start': 8, 'end': 9, 'type': 'time10'},
    'time_slot_4_start': {'start': 9, 'end': 10, 'type': 'time10'},
    'time_slot_4_end': {'start': 10, 'end': 11, 'type': 'time10'},
}

ERROR_STRUCTURE = {
    'number': {'start': 2, 'end': 3, 'type': 'short'},
    'is_ongoing': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 0},
    'is_at_heating_boiler': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 1},
    'is_at_ash_outlet': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 2},
    'is_at_environment': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 4},
    'is_dysfunction': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 5},
    'is_warning': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 6},
    'is_receipted': {'start': 3, 'end': 4, 'type': 'flag', 'bit': 7},
    'status': {'start': 4, 'end': 5, 'type': 'short'},
    'datetime': {'start': 5, 'end': 11, 'type': 'datetime'},
    'text': {'start': 11, 'end': None, 'type': 'string'},
    'status_name': {'start': 4, 'end': 5, 'type': 'short', 'reference': ERROR_STATE},
    'status_name_local': {'start': 4, 'end': 5, 'type': 'short', 'reference': ERROR_STATE_LOCAL},
}

AVAILABLE_VALUE_STRUCTURE = {
    'factor': {'start': 1, 'end': 3, 'type': 'short'},
    'unit': {'start': 6, 'end': 7, 'type': 'string'},
    'address': {'start': 7, 'end': 9, 'type': 'bytes'},
    'text': {'start': 9, 'end': -1, 'type': 'string'},
}


FORCE_MODE_STRUCTURE = {
    'is_force_active': {'start': 0, 'end': 1, 'type': 'flag', 'bit': 0},
}


#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import random
import string

from s3200 import const
from collections import OrderedDict
from datetime import datetime, time
import logging

#---LOGGING---
# create logger with 'spam_application'
logger = logging.getLogger('s3200')
logger.setLevel(logging.WARN)
# create file handler which logs even debug messages
#fh = logging.FileHandler('input_output.log')
#fh.setLevel(logging.INFO)
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
# create formatter and add it to the handlers
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
#fh.setFormatter(formatter)
ch.setFormatter(formatter)
# add the handlers to the logger
#logger.addHandler(fh)
logger.addHandler(ch)


#---HELPER METHODS---
def calculate_checksum(data_bytes: bytes):
    """ Calculates the checksum of the given data.

    :param data_bytes:
    """
    crc = 0x00

    for aByte in data_bytes:
        temp = (aByte * 2) & 0xFF
        crc = crc ^ aByte ^ temp

    return bytes([crc])


def convert_bytes_to_datedaytime(date_bytes: bytes):
    """ Get a date object from its 7byte representation.

    :param date_bytes: the 7 byte long representation of a date

        Form:
        seconds minutes hours day month weekday year

        Example:
        00 1F 12 15 0B 07 0A
        Time: 00 1F 12 -> 18:31:00
        Date: 15 0B 07 0A -> 21.11.; 7. Day of week = Sunday ; (20)10
    """

    date_array = const.StructDateDayTime.unpack(date_bytes)

    second = date_array[0]
    minute = date_array[1]
    hour = date_array[2]
    day = date_array[3]
    month = date_array[4]
    # weekday = date_array[5] # 1=Monday 7=Sunday
    year = 2000 + date_array[6]  # 2000 + byte

    return datetime(year, month, day, hour, minute, second)


def get_bytes_from_date_day_time(datetime_to_set: datetime):
    """ Get the 7byte representation of a date object.
    """

    date_list = []

    date_list.append(datetime_to_set.second)
    date_list.append(datetime_to_set.minute)
    date_list.append(datetime_to_set.hour)
    date_list.append(datetime_to_set.day)
    date_list.append(datetime_to_set.month)
    date_list.append(datetime_to_set.isoweekday())  # 1=Monday 7=Sunday
    date_list.append(datetime_to_set.year - 2000)  # 2000 + byte

    date_bytes = const.StructDateDayTime.pack(date_list)

    return date_bytes

def convert_byte_to_datetime(date_bytes: bytes):
    """ Get a date object from its 6byte representation.

    :param date_bytes: the 6 byte long representation of a date

        Form:
        seconds minutes hours day month weekday year

        Example:
        00 1F 12 15 0B 07 0A
        Time: 00 1F 12 -> 18:31:00
        Date: 15 0B 0A -> 21.11.(20)10
    """

    date_array = const.StructDateTime.unpack(date_bytes)

    second = date_array[0]
    minute = date_array[1]
    hour = date_array[2]
    day = date_array[3]
    month = date_array[4]
    year = 2000 + date_array[5]  # 2000 + byte

    return datetime(year, month, day, hour, minute, second)


def convert_short_to_integer(short_bytes: bytes):
    """ Get a integer of its 2byte representation.

    :param short_bytes: two bytes representing a short
    """

    # allow to convert single byte values
    if len(short_bytes) == 1:
        short_bytes = b'\x00' + short_bytes

    if len(short_bytes) != 2:
        raise ShortUnpackError("2bytes input needed {0}bytes given".format(len(short_bytes)))

    temp_array = const.StructShort.unpack(short_bytes)
    integer = temp_array[0]

    return integer


def convert_integer_to_short(integer: int):
    """ Get the 2byte representation of an integer.

    :param integer: an int to convert to a short. Max. 65025
    """

    if integer < 0 or integer > 65025:
        raise ShortPackError("Given value: {0} could not be represented as 2byte short. " +
                             "Allowed: 0 < value < 65025".format(integer))

    temp_bytes = const.StructShort.pack(integer)

    return temp_bytes


def convert_bytes_to_string(byte_data: bytes):
    """ String from bytes.

    :param byte_data: data to convert
    """

    string_data = byte_data.decode(encoding='cp1252')
    return string_data


def convert_string_to_bytes(string_data: str):
    """ Bytes from string.

    :param string_data: data to convert
    """

    byte_data = string_data.encode(encoding='cp1252')
    return byte_data


def convert_bytes_to_hex(byte_data: bytes):
    """ Convert bytes to a hex string of format 0A FD C3.

    :param byte_data: the byte data to convert to hex
    """
    return ' '.join(['{:02x}'.format(i) for i in byte_data])


def get_flag_from_bytes(data_bytes: bytes, position: int):
    return is_flag_set(data_bytes, position_from_left=position)


def convert_bytes_to_error(error_bytes: bytes):
    """ Get an error dict from bytes representation. """

    return_dict = convert_structure_to_dict(error_bytes, const.ERROR_STRUCTURE)

    return return_dict


def convert_bytes_to_menu_item(menu_item_bytes: bytes):

    # load the definition (what is where) from the constants
    address_def = const.MENU_ITEM_STRUCTURE['address']
    text_def = const.MENU_ITEM_STRUCTURE['text']

    # get the data from their position in the byte data
    address = menu_item_bytes[address_def['start']:address_def['end']]
    text = convert_bytes_to_string(menu_item_bytes[text_def['start']:text_def['end']])

    #build the return dict
    return_dict = {
        'address': address,
        'text': text,
    }

    return return_dict


def convert_bytes_to_configuration(configuration_bytes: bytes):
    return convert_structure_to_dict(configuration_bytes, const.CONFIGURATION_STRUCTURE)



def convert_structure_to_dict(data_bytes, configuration_dict):

    logger.debug('Loaded structure dict: ' + str(configuration_dict) + ' for usage on: ' + str(data_bytes))

    return_dict = {}
    for structure_name, structure_data in configuration_dict.items():

        assert isinstance(structure_data, dict)

        logger.debug('LoadStructure: ' + structure_name )

        structure_type = structure_data['type']
        structure_bytes = data_bytes[structure_data['start']:structure_data['end']]

        logger.debug('Structure extracted bytes: ' + str(structure_bytes))
        value = None

        if structure_type == 'short':
            value = convert_short_to_integer(structure_bytes)

        elif structure_type == 'bytes':
            value = bytes(structure_bytes)

        elif structure_type == 'datetime':
            value = convert_byte_to_datetime(structure_bytes)

        elif structure_type == 'string':
            value = str(convert_bytes_to_string(structure_bytes))

        elif structure_type == 'flag':
            value = bool(get_flag_from_bytes(structure_bytes, structure_data['bit']))

        elif structure_type == 'time10':
            value = convert_bytes_to_time10(structure_bytes)

        # if it is a reference the value should come from the referenced item
        if 'reference' in structure_data:
            reference_dict = structure_data['reference']
            try:
                value = reference_dict[value]
            except KeyError as e:
                raise ReferenceValueNotInConst(str(e)) from e

        logger.debug('Structure converted data: ' + str(value))

        return_dict[structure_name] = value

    return return_dict


def convert_bytes_to_time10(data_bytes):
    data_int = convert_short_to_integer(data_bytes)  # 96h = 150d 0 15:00

    if data_int == 255:  # FFh -> no time set
        return None
    elif data_int > 240:
        raise InvalidValueError('Time value must be between 0 and 240')

    data_str = str(data_int)  # "150"

    hours = int(data_str[:2])  # 15
    minutes = int(data_str[2:]) * 10  # 0 -> time10 has only 10minutes resolution

    return_value = time(hours, minutes)
    return return_value


def convert_bytes_to_state_and_mode(state_mode_bytes: bytes):
    state_and_mode = convert_structure_to_dict(state_mode_bytes, const.STATE_AND_MODE_STRUCTURE)
    state_and_mode_string = state_and_mode['text']

    state_mode_tuple = state_and_mode_string.split(';')
    mode = state_mode_tuple[0]
    state = state_mode_tuple[1]

    return state, mode


def convert_bytes_to_state(state_mode_bytes: bytes):
    state, mode = convert_bytes_to_state_and_mode(state_mode_bytes)
    return state


def convert_bytes_to_mode(state_mode_bytes: bytes):
    state, mode = convert_bytes_to_state_and_mode(state_mode_bytes)
    return mode


def convert_bytes_to_setting(payload: bytes):
    return convert_structure_to_dict(payload, const.SETTING_STRUCTURE)


def convert_bytes_to_available_value(payload: bytes):
    return convert_structure_to_dict(payload, const.AVAILABLE_VALUE_STRUCTURE)

def convert_time_slot_to_bytes(item, weekday, time_slot_1_start, time_slot_1_end, time_slot_2_start, time_slot_2_end,
                             time_slot_3_start, time_slot_3_end, time_slot_4_start, time_slot_4_end):

    raise NotImplementedError
    #return data_bytes

def is_flag_set(flag_data_bytes: bytes, position_from_left):
    bin_string = ""
    for byte in flag_data_bytes:
        temp = bin(byte)[2:]  # cut off 0b prefix
        temp = '00000000' + temp  # fill up with 0
        temp = temp[-8:]  # get the last 8
        bin_string += temp

    #print('bin_string len: ' + str(len(bin_string)) + ' asked position: ' + str(position_from_left))
    return bool(bin_string[position_from_left] == '1')


def escape(data: bytes):
    """ Escape the given data according to the s3200 documentation.

    :param data: the bytes to escape
    """
    return replace_all(data, const.ESCAPE_LIST)


def unescape(data):
    """ Unescape the given data according to the s3200 documentation.

    :param data: the bytes to unescape
    """

    return replace_all(data, const.UNESCAPE_LIST)


def replace_all(data: bytes, dic: OrderedDict):
    """ Replaces all occurrences of a byte.

    :param data: The original bytes
    :param dic: The OrderedDict with the original and the replace values
    """

    for i, j in dic.items():
        data = data.replace(i, j)
    return data


def get_random_string(size=10, chars=string.ascii_letters + string.digits):
    """ Get random string.

    :param size: length of the random string
    :param chars: chars the random string should contain default is ascii_letters + digits
    """
    return ''.join(random.choice(chars) for _ in range(size))


class Frame(object):
    """ An class representing a s3200 communication frame.

    Example:
    To get the operating hours the following frame must be send:
    02 FD 00 03 30 00 62 F2
    02 FD -> Frame start sequence
    00 03 -> Amount of bytes (command + payload)
    30    -> Command (in this case the command for 'get value')
    00 62 -> Payload (in this case the address of the operating hours)
    F2    -> Checksum

    """

    #---CONSTRUCTORS---
    def __init__(self, command: bytes=None, payload: bytes=None):
        self.command = command
        self.payload = payload

        if self.payload is None:
            self.payload = b''

        if self.command is None or len(self.command) < 1:
            raise FrameSyntaxError("Frame must have a command.")

    @staticmethod
    def from_bytes(frame_bytes: bytes):
        """ Get a frame object from its byte representation.

        :param frame_bytes: a frame in byte form
        """

        #example frame (in hex) 02 fd 00 03 30 00 30 04
        start_bytes = frame_bytes[:2]  # 02 fd

        #unescape
        escaped_content = frame_bytes[2:]  # 00 03 30 00 30 04
        unescaped_content = unescape(escaped_content)  # 00 03 30 00 30 04 this example has no escaped content

        #get values from unescaped
        length_bytes = unescaped_content[:2]  # 00 03
        command_byte = unescaped_content[2:3]  # 30
        payload_bytes = unescaped_content[3:-1]  # 00 30
        checksum_byte = unescaped_content[-1:]  # 04

        #check start bytes
        if not start_bytes == const.START_BYTES:
            raise CommunicationError("Start bytes must be: {0} but are: {1}".format(
                convert_bytes_to_hex(const.START_BYTES),
                convert_bytes_to_hex(start_bytes)))

        #check checksum
        calculated_checksum = calculate_checksum(start_bytes + length_bytes + command_byte + payload_bytes)

        #if len(payload_bytes) > 0:
        if not checksum_byte[0] == calculated_checksum[0]:
            raise CommunicationError(
                "Checksum byte doesnt match. Received:{0} Calculated:{1}. Complete frame:{2}".format(
                    convert_bytes_to_hex(checksum_byte), convert_bytes_to_hex(calculated_checksum),
                    convert_bytes_to_hex(frame_bytes)))

        #build return frame
        return Frame(command_byte, payload_bytes)

    #---METHODS---
    def to_bytes(self):
        """ Convert a frame object to its send ready byte representation. """

        #None should be handled like no byte
        if self.payload is None:
            self.payload = b''

        #calculate length
        length = len(self.command) + len(self.payload)

        length_bytes = convert_integer_to_short(length)

        #calculate checksum
        checksum = calculate_checksum(const.START_BYTES + length_bytes + self.command + self.payload)

        #escape all except the start bytes
        content_unescaped = length_bytes + self.command + self.payload + checksum
        content_escaped = escape(content_unescaped)

        #build final frame
        frame = const.START_BYTES + content_escaped

        return frame

    def __str__(self):
        return "<Frame command:{0} payload:{1}>".format(convert_bytes_to_hex(self.command),
                                                        convert_bytes_to_hex(self.payload))


class S3200Error(Exception):
    """Base class for exceptions in this module.

       Attributes:
        msg  -- explanation of the error
    """

    def __init__(self, msg=''):
        self.msg = msg


class ShortPackError(S3200Error):
    """Exception raised for values that would not fit into the 2byte short."""


class ShortUnpackError(S3200Error):
    """Exception raised for values that would not fit into the 2byte short."""


class FrameSyntaxError(S3200Error):
    """Exception raised when the syntax or value of a frame is not correct."""


class ValueNotDefinedError(S3200Error):
    """Exception raised when the value asked for is not defined id address_dict."""


class CommandNotDefinedError(S3200Error):
    """Exception raised when the command asked for is not defined id command_dict."""


class CommunicationError(S3200Error):
    """Exception raised when the command asked for is not defined id command_dict."""


class ReadonlyError(S3200Error):
    """Exception raised when trying to write a value when S3200 object is in readonly mode."""


class InvalidValueError(S3200Error):
    """ Exception raised when an unusable value is returned.
    """


class ValueSetError(S3200Error):
    """ Exception raised when an error occurred during setting a value
    """


class NothingToReadError(S3200Error):
    """ Exception raised when an error occurred during setting a value
    """

class WrongNumberOfAnswerFramesError(S3200Error):
    """ Exception raised when an error occurred during setting a value
    """
    def __init__(self, msg, expected, got, base_error):
        super().__init__(msg)
        self.expected = expected
        self.got = got
        self.base_error = base_error

class ReferenceValueNotInConst(S3200Error):
    """ Exception raised when value is references that is not defined in const
    """


#!/usr/bin/python3
# -*- coding: UTF-8 -*-
from unittest import TestCase
from s3200 import const
from s3200.obj import S3200
from s3200 import core
import datetime


class TestS3200(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.s = S3200('dummy',
                             value_definitions=const.VALUE_DEFINITIONS)

    def test_get_value(self):  #ok
        t = self.s.get_value('residual_oxygen')
        #print(t)
        self.assertEquals(432.2, t)
        #print(str(t[0]) + ": " + str(t[1]))

    def test_test_connection(self):  #ok
        self.assertTrue(self.s.test_connection())

    def test_get_datetime(self): #ok
        self.assertEquals(datetime.datetime(2010, 11, 21, 18, 31), self.s.get_datetime())

    def test_get_version(self):  #ok '50.04.05.09'
        self.assertEquals('50.04.04.14', self.s.get_version())

    def test_get_errors(self):  #ok
        #print(str(self.s.get_errors()[0]))
        self.assertDictEqual({'is_at_ash_outlet': True,
                             'is_receipted': False,
                             'number': 109,
                             'is_dysfunction': False,
                             'is_ongoing': True,
                             'is_at_environment': False,
                             'datetime': datetime.datetime(2013, 4, 11, 10, 38, 37),
                             'is_warning': True,
                             'text': 'Zündversuch nicht gelungen von Hand Anheizen!',
                             'is_at_heating_boiler': False,
                             'status': 4,
                             'status_name': 'Gone',
                             'status_name_local': 'Gegangen'}, self.s.get_errors()[0])

    def test_get_state(self):  #ok
        self.assertEqual('STÖRUNG', self.s.get_state())

    def test_get_mode(self):  #ok
        self.assertEqual('Übergangsbetr', self.s.get_mode())

    def test_get_menu(self):  #ok
        item = self.s.get_menu()[0]
        self.assertEquals("Proportionalfaktor des Mischerreglers", item["text"])
        self.assertEquals(b'\x00\x53', item["address"])

    def test_get_setting(self):  #ok
        self.assertEquals(84, self.s.get_setting('heating_boiler_should_temperature'))

    def test_get_setting_info(self):  #ok
        #print("test_get_setting_info: " + str(self.s.get_setting_info('heating_boiler_should_temperature')))
        self.assertDictEqual({'comma': 0,
                              'max_value': 90,
                              'address': bytearray(b'\x00\x1c'),
                              'value': 84,
                              'standard': 80,
                              'factor': 2,
                              'min_value': 70,
                              'unit': '°'}, self.s.get_setting_info('heating_boiler_should_temperature'))

    def test_get_configuration(self):  #ok
        #print("test_get_configuration: " + str(self.s.get_configuration()))
        self.assertDictEqual({'boiler_7': False,
                              'boiler_4': False,
                              'boiler_6': False,
                              'boiler_5': False,
                              'solar_2': False,
                              'boiler_2': False,
                              'boiler_3': False,
                              'heater_circuit_1': True,
                              'boiler_8': False,
                              'solar_1': False,
                              'remote_control_2': False,
                              'heater_circuit_2': True,
                              'remote_control_1': False,
                              'boiler_1': True}, self.s.get_configuration())

    def test_get_digital_input(self):  #no
        #print("test_get_digital_input: " + str(self.s.get_digital_input('door_contact')))
        self.assertEquals(True, self.s.get_digital_input('door_contact'))

    def test_get_digital_output(self):  #no
        #print("test_get_digital_output: " + str(self.s.get_digital_output('heating_circuit_pump_1')))
        self.assertEquals(True, self.s.get_digital_output('heating_circuit_pump_1'))

    def test_get_analog_output(self):
        #print("test_get_analog_output: " + str(self.s.get_analog_output('primary_air')))
        self.assertEquals(99, self.s.get_analog_output('primary_air'))

    def test_get_available_values(self):  #ok
        #print("test_get_available_values: " + str(self.s.get_available_values()))
        self.assertDictEqual({'factor': 2,
                              'text': 'Kesseltemperatur',
                              'address': b'\x00\x00',
                              'unit': '°'}, self.s.get_available_values()[0])

    def test_set_setting(self):
        self.assertRaises(core.ReadonlyError, self.s.set_setting, 'heating_boiler_should_temperature', 42)
        self.s = S3200('dummy', readonly=False)

        self.s.set_setting('heating_boiler_should_temperature', 84)  # 84 is already set
        self.s.set_setting('heating_boiler_should_temperature', 81)

        # TODO repair dummy to enable this test
        #self.assertRaises(core.ValueSetError, self.s.set_setting, 'heating_boiler_should_temperature', 120),

        self.s = S3200('dummy', readonly=True)









import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="s3200",
    version="0.0.1",
    author="cbasix",
    author_email="cbasix@users.noreply.github.com",
    description="A Python package for controlling the Fröling heating system (S3200) over the maintenance interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cbasix/s3200",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


#!/usr/bin/python3
# -*- coding: UTF-8 -*-



from s3200 import const, core
from s3200.core import CommunicationError, Frame
from s3200.test.dummy import DummySerial
import logging

logger = logging.getLogger('s3200')

try:
    from serial import Serial, EIGHTBITS, PARITY_NONE, STOPBITS_ONE
except:
    logger.error("WARN: Could not load serial module. Only dummy mode!")

class Connection(object):
    """ A class representing a serial connection to a s3200 device. """

    def __init__(self, serial_port_name="/dev/ttyAMA0"):
        self.serial_port_name = serial_port_name

    def send(self, command: bytes=None, payload: bytes=None):
        """ Shortcut for send_frame. Builds the Frame object and sends it. """

        frame = Frame(command, payload)
        return_frame = self.send_frame(frame)

        return return_frame

    def open_serial(self):
        """Opens a serial port and returns it."""
        if self.serial_port_name == 'dummy':
            serial_port = DummySerial()
        else:
            serial_port = Serial(self.serial_port_name, 57600, EIGHTBITS, PARITY_NONE, STOPBITS_ONE, timeout=3)
        return serial_port

    def send_frame(self, frame, read_answer_frames=1):
        """ Sends one frame and receives the answer frame

        :param frame: the frame to send
        :return frame: the answer frame
        :raise: different exceptions that could occur during communication
        """

        serial_port = self.open_serial()
        answer_frames = []

        try:
            #send the frame
            logger.debug('sending: ' + str(frame.to_bytes()))
            serial_port.write(frame.to_bytes())

            #read the answer bytes
            for i in range(read_answer_frames):
                answer_bytes = Connection._read_one_frame(serial_port)
                logger.debug('read answer:' + str(answer_bytes))
                #wrong set_setting returns 2x -> logger.info("'"+str(core.convert_bytes_to_hex(frame.to_bytes()))+"'"+': ' + str(answer_bytes)+',')
                #make a frame from the bytes
                answer_frame = Frame.from_bytes(answer_bytes)
                answer_frames.append(answer_frame)

        except core.NothingToReadError as e:
            serial_port.flushInput()
            raise core.WrongNumberOfAnswerFramesError("Got wrong no of answer frames",
                                                      read_answer_frames, len(answer_frames), e)

        finally:
            serial_port.close()

        if len(answer_frames) > 1:
            return answer_frames
        else:
            return answer_frames[0]

    @staticmethod
    def _read_one_frame(serial_port):

        frame_start_bytes = serial_port.read(2)

        if serial_port.inWaiting() == 0:
            raise core.NothingToReadError("No Bytes to Read")

        #get length
        length_bytes = Connection._read_escaped(serial_port, 2)
        unescaped_length_bytes = core.unescape(length_bytes)
        if len(unescaped_length_bytes) != 2:
            raise core.CommunicationError("Didn't get 2 length bytes. Maybe timeout, or other reading error.")

        length = core.convert_short_to_integer(unescaped_length_bytes)

        read_bytes = Connection._read_escaped(serial_port, length + 1)  # +1 for read the checksum too
        complete_frame = frame_start_bytes + length_bytes + read_bytes

        return complete_frame

    @staticmethod
    def _read_escaped(serial_port, length):
        my_byte = bytearray()

        for i in range(length):
            byte = serial_port.read(1)[0]
            my_byte.append(byte)

            if byte in const.ESCAPED_IDENTIFIER:
                second_byte = serial_port.read(1)[0]
                my_byte.append(second_byte)

        return bytes(my_byte)

    def get_list(self, command_start_address: bytes, command_next_address: bytes, max_loops=500):
        """ Get all items of a list """

        output = []
        answer_frame = self.send(command_start_address)

        while answer_frame.payload != b'\x00' :  # TODO add constant
            logger.debug('get_list payload: ' + str(answer_frame.payload))

            if answer_frame.payload == b'\x01' :
                logger.debug('ignore payload: ' + str(answer_frame.payload))
            else:
                output.append(answer_frame)

            answer_frame = self.send(command_next_address, b'\x01')
            logger.debug('get_list len: ' + str(len(output)))

            #prevent endless loops
            if len(output) > max_loops:
                raise ValueError("Reached max_loops: " + str(max_loops))

        return output


#!/usr/bin/python3
# -*- coding: UTF-8 -*-

from datetime import datetime, time
from s3200 import const, core, net
from s3200.net import Frame
import logging


logger = logging.getLogger('s3200')


class S3200(object):
    """ A class representing a s3200 object. """

    def __init__(self, serial_port_name="/dev/ttyAMA0",
                 readonly=True,
                 value_definitions=const.VALUE_DEFINITIONS,
                 setting_definitions=const.SETTING_DEFINITIONS,
                 command_definitions=const.COMMAND_DEFINITIONS,
                 digital_input_definitions=const.DIGITAL_INPUT_DEFINITIONS,
                 digital_output_definitions=const.DIGITAL_OUTPUT_DEFINITIONS,
                 analog_output_definitions=const.ANALOG_OUTPUT_DEFINITIONS,
                 ):

        self.connection = net.Connection(serial_port_name=serial_port_name)
        self.readonly = readonly
        self.value_definitions = value_definitions
        self.setting_definitions = setting_definitions
        self.command_definitions = command_definitions
        self.digital_input_definitions = digital_input_definitions
        self.digital_output_definitions = digital_output_definitions
        self.analog_output_definitions = analog_output_definitions

        #if not (readonly or serial_port_name == 'dummy'):
            #raise NotImplementedError('Currently only readonly mode is supported.')

    def _test_readonly_(self):
        if self.readonly:
            raise core.ReadonlyError("Can not set values in readonly mode.")

    def get_value(self, *args: str, with_local_name: bool=False):
        """ Get value by name.

        :param with_local_name: if yes output is a tuple with (name, value) instead of value only
        :param args: name of the value as specified in address_dict
        """
        return_list = []
        for value_name in args:
            if not self.value_definitions[value_name]:
                raise core.ValueNotDefinedError("Address for value: '{0}' not defined in address_dict".format(value_name))

            if not self.command_definitions['get_value']:
                raise core.CommandNotDefinedError("Address for command: 'get_value' not defined in command_dict")

            value_definition = self.value_definitions[value_name]

            # Prepare the frame and get the answer
            command_address = self.command_definitions['get_value']['address']
            value_address = value_definition['address']

            answer_frame = self.connection.send(command_address, value_address)

            value = core.convert_short_to_integer(answer_frame.payload)
            value = value / value_definition['factor']

            if with_local_name:
                return_list.append((value_definition['local_name'], value))
            else:
                return_list.append(value)

        if len(args) == 1:
            return return_list[0]
        else:
            return return_list


    def test_connection(self):
        """ Tests the connection.

            Tests the connection by sending a random string and reading it back.

            :return: True if connection was successful. False otherwise.
        """

        command_address = self.command_definitions['test_connection']['address']
        random_string = core.get_random_string(15)
        payload = core.convert_string_to_bytes(random_string)

        try:
            answer_frame = self.connection.send(command_address, payload)
        except core.CommunicationError as e:
            return False

        return_string = core.convert_bytes_to_string(answer_frame.payload)

        if return_string == random_string:
            return True

        return False

    def get_version(self):
        """ Gets the software version from the heater.

        :return: A string containing the version
        """

        command_address = self.command_definitions['get_version_and_datetime']['address']
        answer_frame = self.connection.send(command_address)

        #first 4bytes are the software version the rest is for the date
        version_bytes = answer_frame.payload[:4]

        #convert into . separated string
        version_string = '.'.join(['{:02x}'.format(i) for i in version_bytes])
        return version_string

    def get_datetime(self):
        """ Gets the date and time from the heater.

        :return: A datetime object
        """

        command_address = self.command_definitions['get_version_and_datetime']['address']
        answer_frame = self.connection.send(command_address)

        #first 4bytes are the software version the rest is for the date
        date_bytes = answer_frame.payload[4:11]

        #convert into . separated string
        return_date = core.convert_bytes_to_datedaytime(date_bytes)
        return return_date

    def set_datetime(self, datetime_to_set: datetime):
        """ Set the date and time of the heater. """
        self._test_readonly_()

        command_address = self.command_definitions['set_datetime']['address']
        date_bytes = core.get_bytes_from_date_day_time(datetime_to_set)

        answer_frame = self.connection.send(command_address, date_bytes)

        # TODO Test answer frame

    def get_errors(self):
        """ Get all errors currently in the error buffer. """

        command_start_address = self.command_definitions['get_error']['address']
        command_next_address = self.command_definitions['get_next_error']['address']

        output = []
        error_frames = self.connection.get_list(command_start_address, command_next_address)

        for frame in error_frames:
            error = core.convert_bytes_to_error(frame.payload)
            output.append(error)

        return output

    def get_time_slots(self):
        """ Get the currently set time slots. """

        command_start_address = self.command_definitions['get_time_slot']['address']
        command_next_address = self.command_definitions['get_next_time_slot']['address']

        output = []
        time_slots = self.connection.get_list(command_start_address, command_next_address)

        for frame in time_slots:
            time_slot = core.convert_structure_to_dict(frame.payload, const.TIME_SLOT_STRUCTURE)

            output.append(time_slot)

        return output

    def set_time_slot(self,
                      item: str,
                      weekday: int,
                      time_slot_1_start: time,
                      time_slot_1_end: time,
                      time_slot_2_start: time,
                      time_slot_2_end: time,
                      time_slot_3_start: time,
                      time_slot_3_end: time,
                      time_slot_4_start: time,
                      time_slot_4_end: time,
                      ):
        """ Set a timeslot. NOT IMPLEMENTED YET. """

        self._test_readonly_()
        raise NotImplementedError()

        data_bytes = core.convert_time_slot_to_bytes(item,
                                                     weekday,
                                                     time_slot_1_start,
                                                     time_slot_1_end,
                                                     time_slot_2_start,
                                                     time_slot_2_end,
                                                     time_slot_3_start,
                                                     time_slot_3_end,
                                                     time_slot_4_start,
                                                     time_slot_4_end,)

        command_address = self.command_definitions['set_time_slot']['address']
        frame = Frame(command_address, data_bytes)
        answer_frame = self.connection.send_frame(frame)

        # TODO Check answer frame
        #if not answer_frame == frame:
        #    raise core.ValueSetError('Error occurred while setting value.')

    def get_configuration(self):
        """ Get the active and connected boilers, heating circuits and solar. """

        command_address = self.command_definitions['get_configuration']['address']
        answer_frame = self.connection.send(command_address)

        return_dict = core.convert_bytes_to_configuration(answer_frame.payload)

        return return_dict

    def get_state(self):
        """ Get the state of the heater.

         Example: Heizen(Heating) """

        command_address = self.command_definitions['get_heater_state_and_mode']['address']
        answer_frame = self.connection.send(command_address)

        state = core.convert_bytes_to_state(answer_frame.payload)

        return state

    # TODO better docstrings
    def get_mode(self):
        """ Get the mode of the heater.

         Example: Übergangsbetr """
        command_address = self.command_definitions['get_heater_state_and_mode']['address']
        answer_frame = self.connection.send(command_address)

        mode = core.convert_bytes_to_mode(answer_frame.payload)

        return mode

    def get_menu(self):
        """Get the complete menu structure. """

        command_start_address = self.command_definitions['get_menu_item']['address']
        command_next_address = self.command_definitions['get_next_menu_item']['address']

        output = []
        error_frames = self.connection.get_list(command_start_address, command_next_address, max_loops=5000)

        for frame in error_frames:
            error = core.convert_bytes_to_menu_item(frame.payload)
            output.append(error)

        return output

    def get_available_values(self):
        """Get all available values from the heater. """

        command_start_address = self.command_definitions['get_available_value']['address']
        command_next_address = self.command_definitions['get_next_available_value']['address']

        output = []
        available_value_frames = self.connection.get_list(command_start_address, command_next_address)

        for frame in available_value_frames:
            available_value = core.convert_structure_to_dict(frame.payload, const.AVAILABLE_VALUE_STRUCTURE)
            output.append(available_value)

        return output

    def get_setting(self, setting_name):
        """Get the specified setting from the heater. """

        command_address = self.command_definitions['get_setting']['address']
        value_address = self.setting_definitions[setting_name]['address']

        answer_frame = self.connection.send(command_address, value_address)
        setting = core.convert_bytes_to_setting(answer_frame.payload)

        return_value = setting['value'] / setting['factor']

        if setting['comma'] == 0:
            return_value = int(return_value)

        return return_value

    def get_setting_info(self, setting_name):
        """Get the specified setting value, min_value, max_value, standard_value and others. """

        command_address = self.command_definitions['get_setting']['address']
        value_address = self.setting_definitions[setting_name]['address']

        answer_frame = self.connection.send(command_address, value_address)
        setting = core.convert_bytes_to_setting(answer_frame.payload)

        setting['value'] = setting['value'] / setting['factor']

        if setting['comma'] == 0:
            setting['value'] = int(setting['value'])

        return setting

    def set_setting(self, setting_name: str, value: int):
        """Set the specified setting to the given value."""
        self._test_readonly_()

        old_value = self.get_setting(setting_name)
        if old_value == value:
            return

        command_address = self.command_definitions['set_setting']['address']
        value_address = self.setting_definitions[setting_name]['address']
        factor = self.setting_definitions[setting_name]['factor']

        value = value * factor
        payload = value_address + core.convert_integer_to_short(value)
        frame = Frame(command_address, payload)
        #print(core.convert_bytes_to_hex(frame.to_bytes()))

        # TODO make better
        try:
            answer_frames = self.connection.send_frame(frame, read_answer_frames=2)
        except core.WrongNumberOfAnswerFramesError as e:
            if e.got == 1:
                raise core.ValueSetError("Setting could not be set. Maybe the value you want to set is out of range? "
                                     "Use get_setting_info to check.")
            else:
                raise e.base_error
        assert(isinstance(answer_frames[0], Frame))
        assert(isinstance(answer_frames[1], Frame))
        #TODO WORK HERE

        if str(frame) != str(answer_frames[0]) or str(frame) != str(answer_frames[1]):
            raise core.ValueSetError("Setting could not be set. Heater returned different values")

    def get_digital_input(self, input_name):
        """Get the state of a digital input."""

        command_address = self.command_definitions['get_digital_input']['address']
        value_address = self.digital_input_definitions[input_name]

        answer_frame = self.connection.send(command_address, value_address)
        digital_input = core.convert_structure_to_dict(answer_frame.payload, const.DIGITAL_INPUT_STRUCTURE)

        return_bool = None
        if digital_input['mode'] == 'A':  # \x41 = A = Auto
            return_bool = bool(digital_input['value'] == 1)
        elif digital_input['mode'] == '0':  # \x30 = 0 = False
            return_bool = False
        elif digital_input['mode'] == '1':  # \x31 = 1 = True
            return_bool = True

        return return_bool

    def get_digital_output(self, output_name):
        """Get the state of a digital output."""

        command_address = self.command_definitions['get_digital_output']['address']
        value_address = self.digital_output_definitions[output_name]

        answer_frame = self.connection.send(command_address, value_address)
        digital_output = core.convert_structure_to_dict(answer_frame.payload, const.DIGITAL_OUTPUT_STRUCTURE)

        return_bool = None
        if digital_output['mode'] == 'A':  # \x41 = A = Auto
            return_bool = bool(digital_output['value'] == 1)
        elif digital_output['mode'] == '0':  # \x30 = 0 = False
            return_bool = False
        elif digital_output['mode'] == '1':  # \x31 = 1 = True
            return_bool = True

        return return_bool

    def get_analog_output(self, output_name):
        """Get the state of a analog output."""
        command_address = self.command_definitions['get_analog_output']['address']
        value_address = self.analog_output_definitions[output_name]

        answer_frame = self.connection.send(command_address, value_address)
        analog_output = core.convert_structure_to_dict(answer_frame.payload, const.ANALOG_OUTPUT_STRUCTURE)

        if analog_output['mode'] == 255:  # Auto mode
            return analog_output['value']
        else:
            return analog_output['mode']  # Manual override

    # def set_force_mode(self, is_force_mode: bool):
    #     """ Sets the Force mode.
    #
    #     Force mode is a testing mode. When force mode is active you can override the input and output values
    #     of the heater manually.  If the heater receives no command within 30 seconds the force mode gets
    #     deactivated automatically.
    #     """
    #
    #     self._test_readonly_()
    #     raise NotImplementedError()
    #
    #     # TODO implement enter force mode
    #
    # def is_force_active(self):
    #     command_address = self.command_definitions['get_force_mode']['address']
    #
    #     answer_frame = self.connection.send(command_address)
    #
    #     result_dict = core.convert_structure_to_dict(answer_frame.payload, const.FORCE_MODE_STRUCTURE)
    #
    #     return result_dict['is_force_active']
    #
    #
    # def set_digital_input(self, input_name, value: bool):
    #     self._test_readonly_()
    #
    #     assert(isinstance(bool, value))
    #     raise NotImplementedError()
    #
    #     command_address = self.command_definitions['set_digital_input']['address']
    #     value_address = self.digital_input_definitions[input_name]
    #
    #     if value:
    #         value = b'\x01'
    #     else:
    #         value = b'\x00'
    #
    #     answer_frame = self.connection.send(command_address, value_address + value)
    #
    #     # TODO test answer frame
    #
    # def set_digital_output(self, output_name, value):
    #     self._test_readonly_()
    #
    #     assert(isinstance(bool, value))
    #     raise NotImplementedError()
    #
    #     command_address = self.command_definitions['set_digital_output']['address']
    #     value_address = self.digital_output_definitions[output_name]
    #
    #     if value:
    #         value = b'\x01'
    #     else:
    #         value = b'\x00'
    #
    #     answer_frame = self.connection.send(command_address, value_address + value)
    #
    #     #TODO test answer frame
    #
    # def set_analog_output(self, output_name, value):
    #     self._test_readonly_()
    #
    #     assert(isinstance(int, value))
    #     raise NotImplementedError()
    #
    #     command_address = self.command_definitions['set_digital_input']['address']
    #     value_address = self.digital_output_definitions[output_name]
    #
    #     value = core.convert_integer_to_short(value)
    #
    #     answer_frame = self.connection.send(command_address, value_address + value)
    #
    #     #TODO test answer frame